from dataclasses import dataclass
from typing import Optional, Tuple

from fretboardgtr.constants import BLACK, WHITE

TEXT_OFFSET = "0.3em"
TEXT_STYLE = "text-anchor:middle"
import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.elements.base import FretBoardElement


@dataclass
class OpenNoteConfig(ConfigIniter):
    """OpenNote element configuration."""

    radius: int = 20
    color: str = WHITE
    stroke_color: str = BLACK
    stroke_width: int = 3
    text_color: str = BLACK
    fontsize: int = 20
    fontweight: str = "bold"


class OpenNote(FretBoardElement):
    """Open notes elements to be drawn in the final fretboard."""

    def __init__(
        self,
        name: str,
        position: Tuple[float, float],
        config: Optional[OpenNoteConfig] = None,
    ):
        self.config = config if config else OpenNoteConfig()
        self.name = name
        self.x = position[0]
        self.y = position[1]

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the OpenNote to a svgwrite object.

        This maps the OpenNoteConfig configuration attributes to the svg
        attributes
        """
        note = svgwrite.container.Group()
        circle = svgwrite.shapes.Circle(
            (self.x, self.y),
            r=self.config.radius,
            fill=self.config.color,
            stroke=self.config.stroke_color,
            stroke_width=self.config.stroke_width,
        )
        text = svgwrite.text.Text(
            self.name,
            insert=(self.x, self.y),
            dy=[TEXT_OFFSET],
            font_size=self.config.fontsize,
            fill=self.config.text_color,
            font_weight=self.config.fontweight,
            style=TEXT_STYLE,
        )
        note.add(circle)
        note.add(text)
        return note


@dataclass
class FrettedNoteConfig(ConfigIniter):
    radius: int = 20
    color: str = WHITE
    stroke_color: str = BLACK
    stroke_width: int = 3
    text_color: str = BLACK
    fontsize: int = 20
    fontweight: str = "bold"


class FrettedNote(FretBoardElement):
    """Fretted notes elements to be drawn in the final fretboard."""

    def __init__(
        self,
        name: str,
        position: Tuple[float, float],
        config: Optional[FrettedNoteConfig] = None,
    ):
        self.config = config if config else FrettedNoteConfig()
        self.name = name
        self.x = position[0]
        self.y = position[1]

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the FrettedNote to a svgwrite object.

        This maps the FrettedNoteConfig configuration attributes to the
        svg attributes
        """
        note = svgwrite.container.Group()
        circle = svgwrite.shapes.Circle(
            (self.x, self.y),
            r=self.config.radius,
            fill=self.config.color,
            stroke=self.config.stroke_color,
            stroke_width=self.config.stroke_width,
        )

        text = svgwrite.text.Text(
            self.name,
            insert=(self.x, self.y),
            dy=[TEXT_OFFSET],
            font_size=self.config.fontsize,
            fill=self.config.text_color,
            font_weight="bold",
            style=TEXT_STYLE,
        )
        note.add(circle)
        note.add(text)
        return note



from fretboardgtr.elements.background import Background, BackgroundConfig


def test_background_get_svg():
    background = Background(position=(0.0, 0.0), size=(10.0, 10.0))
    attribs = background.get_svg().attribs
    assert attribs["x"] == 0.0
    assert attribs["y"] == 0.0
    assert attribs["width"] == 10.0
    assert attribs["height"] == 10.0


def test_background_get_svg_custom_config():
    background_config = BackgroundConfig(color="blue")
    background = Background(
        position=(0.0, 0.0), size=(10.0, 10.0), config=background_config
    )
    attribs = background.get_svg().attribs
    assert attribs["x"] == 0.0
    assert attribs["y"] == 0.0
    assert attribs["width"] == 10.0
    assert attribs["height"] == 10.0
    assert attribs["fill"] == "blue"



from dataclasses import dataclass
from typing import Optional, Tuple

import svgwrite

from fretboardgtr.base import ConfigIniter
from fretboardgtr.constants import GRAY
from fretboardgtr.elements.base import FretBoardElement

TEXT_OFFSET = "0.3em"
TEXT_STYLE = "text-anchor:middle"


@dataclass
class TuningConfig(ConfigIniter):
    """Tuning element configuration."""

    color: str = GRAY
    fontsize: int = 20
    fontweight: str = "normal"


class Tuning(FretBoardElement):
    """Tuning texts elements to be drawn in the final fretboard."""

    def __init__(
        self,
        name: str,
        position: Tuple[float, float],
        config: Optional[TuningConfig] = None,
    ):
        self.config = config if config else TuningConfig()
        self.name = name
        self.x = position[0]
        self.y = position[1]

    def get_svg(self) -> svgwrite.base.BaseElement:
        """Convert the Tuning to a svgwrite object.

        This maps the TuningConfig configuration attributes to the svg
        attributes
        """
        text = svgwrite.text.Text(
            self.name,
            insert=(self.x, self.y),
            dy=[TEXT_OFFSET],
            fill=self.config.color,
            font_size=self.config.fontsize,
            font_weight=self.config.fontweight,
            style=TEXT_STYLE,
        )
        return text



"""Top-level package for FretBoardGtr."""
from fretboardgtr._version import version_str
from fretboardgtr.elements.background import Background, BackgroundConfig
from fretboardgtr.elements.fret_number import FretNumber, FretNumberConfig
from fretboardgtr.elements.frets import Fret, FretConfig
from fretboardgtr.elements.neck_dots import NeckDot, NeckDotConfig
from fretboardgtr.elements.notes import (
    FrettedNote,
    FrettedNoteConfig,
    OpenNote,
    OpenNoteConfig,
)
from fretboardgtr.elements.nut import Nut, NutConfig
from fretboardgtr.elements.strings import String, StringConfig
from fretboardgtr.elements.tuning import Tuning, TuningConfig
from fretboardgtr.fretboard import FretBoard, VerticalFretBoard
from fretboardgtr.fretboards.base import FretBoardLike
from fretboardgtr.fretboards.config import FretBoardConfig, FretBoardGeneralConfig
from fretboardgtr.fretboards.converters import FretBoardToSVGConverter
from fretboardgtr.fretboards.elements import FretBoardElements
from fretboardgtr.fretboards.fretboard import FretBoardContainer
from fretboardgtr.fretboards.vertical import VerticalFretBoardContainer
from fretboardgtr.note_colors import NoteColors
from fretboardgtr.notes_creators import NotesContainer

__author__ = "Antoine Gibek"
__email__ = "antoine.gibek@gmail.com"
__version__ = version_str



import os
from pathlib import Path

from fretboardgtr.fretboard import FretBoard, VerticalFretBoard
from fretboardgtr.notes_creators import ScaleFromName

C_IONIAN_SVG = Path(__file__).parent / "data" / "c_ionian.svg"
OUTPUTS_ARTIFACT_FOLDER = Path(__file__).parent / "data" / "outputs" / "e2e"


def remove_test_file(path: Path):
    if path.exists():
        os.remove(str(path))


def test_c_major_fretboard():
    fretboard = FretBoard()
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "horizontal.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()


def test_c_major_vertical_fretboard():
    fretboard = VerticalFretBoard()
    c_major = ScaleFromName(root="C", mode="Ionian").get()
    fretboard.add_notes(scale=c_major)
    os.makedirs(OUTPUTS_ARTIFACT_FOLDER, exist_ok=True)
    tmp_file = OUTPUTS_ARTIFACT_FOLDER / "vertical.svg"
    remove_test_file(tmp_file)
    assert not tmp_file.exists()
    fretboard.export(tmp_file)
    assert tmp_file.exists()



import copy
from typing import List, Optional, Tuple, Union

from fretboardgtr.constants import STANDARD_TUNING
from fretboardgtr.elements.background import Background
from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.elements.cross import Cross
from fretboardgtr.elements.fret_number import FretNumber
from fretboardgtr.elements.frets import Fret
from fretboardgtr.elements.neck_dots import NeckDot
from fretboardgtr.elements.notes import FrettedNote, OpenNote
from fretboardgtr.elements.nut import Nut
from fretboardgtr.elements.strings import String
from fretboardgtr.elements.tuning import Tuning
from fretboardgtr.fretboards.base import FretBoardLike
from fretboardgtr.fretboards.config import FretBoardConfig
from fretboardgtr.fretboards.elements import FretBoardElements
from fretboardgtr.notes_creators import NotesContainer
from fretboardgtr.utils import (
    chromatic_position_from_root,
    get_note_from_index,
    get_valid_dots,
    note_to_interval_name,
)


class FretBoardContainer(FretBoardLike):
    """Class containing the different elements of a fretboard.

    Also contains associated method to add some.
    """

    def __init__(
        self,
        tuning: Optional[List[str]] = None,
        config: Optional[FretBoardConfig] = None,
    ):
        self.config = config if config is not None else FretBoardConfig()
        self.tuning = tuning if tuning is not None else STANDARD_TUNING
        self.elements = FretBoardElements()
        self.init()

    def set_config(self, config: FretBoardConfig) -> None:
        self.config = config

    def init(self) -> None:
        """Init the fretboard by adding essential elements.

        This adds :
            background
            fret_numbers
            neck_dots
            frets
            nut
            tuning
            strings
        """
        self.add_background()
        self.add_fret_numbers()
        self.add_neck_dots()
        self.add_frets()
        self.add_nut()
        self.add_tuning()
        self.add_strings()

    def add_background(self) -> None:
        """Build and add background element."""
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        open_fret_width = self.config.general.fret_width
        x = self.config.general.x_start + open_fret_width
        y = self.config.general.y_start
        width = (number_of_frets) * (self.config.general.fret_width)
        height = (len(self.tuning) - 1) * self.config.general.fret_height
        background = Background((x, y), (width, height), config=self.config.background)
        self.elements.background = background

    def add_neck_dots(self) -> None:
        """Build and add neck dot elements."""
        dots = get_valid_dots(
            self.config.general.first_fret, self.config.general.last_fret
        )
        for dot in dots:
            x = (
                self.config.general.x_start
                + (0.5 + dot - self.config.general.first_fret)
                * self.config.general.fret_width
            )
            y = (
                self.config.general.y_start
                + (len(self.tuning) / 2 - (1 / 2)) * self.config.general.fret_height
            )
            if dot % 12 == 0:
                # Add two dots dot is multiple of 12
                lower_position = (
                    x,
                    y - self.config.general.fret_height,
                )
                upper_position = (
                    x,
                    y + self.config.general.fret_height,
                )
                upper_dot = NeckDot(upper_position, config=self.config.neck_dots)
                lower_dot = NeckDot(lower_position, config=self.config.neck_dots)
                self.elements.neck_dots.append(upper_dot)
                self.elements.neck_dots.append(lower_dot)
            else:
                center_position = (
                    x,
                    y,
                )
                center_dot = NeckDot(center_position, config=self.config.neck_dots)
                self.elements.neck_dots.append(center_dot)

    def add_frets(self) -> None:
        """Build and add fret elements."""
        show_nut = self.config.general.show_nut
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret

        # Skip the open virtual fret
        first_fret = 1
        if self.config.general.first_fret == 0 and show_nut:
            first_fret = 2

        # +2 is to close the last fret of fretboard
        for fret_no in range(first_fret, number_of_frets + 2):
            x = self.config.general.x_start + (self.config.general.fret_width) * (
                fret_no
            )
            y_start = self.config.general.y_start
            y_end = self.config.general.y_start + (self.config.general.fret_height) * (
                len(self.tuning) - 1
            )

            fret = Fret((x, y_start), (x, y_end), config=self.config.frets)
            self.elements.frets.append(fret)

    def add_strings(self) -> None:
        """Build and add string elements."""
        # begin before if min fret !=0 because no open chords
        open_fret_width = self.config.general.fret_width
        x_start = self.config.general.x_start + open_fret_width
        x_end = self.config.general.x_start + (
            self.config.general.fret_width
            + (self.config.general.last_fret - self.config.general.first_fret)
            * self.config.general.fret_width
        )

        for i, _ in enumerate(self.tuning):
            start = (
                x_start,
                self.config.general.y_start + (self.config.general.fret_height) * (i),
            )
            end = (
                x_end,
                self.config.general.y_start + (self.config.general.fret_height) * (i),
            )
            string = String(start, end, config=self.config.strings)
            self.elements.strings.append(string)

    def add_nut(self) -> None:
        """Build and add nut element."""
        if self.config.general.first_fret != 0 or not self.config.general.show_nut:
            return None
        open_fret_width = self.config.general.fret_width

        start = (
            self.config.general.x_start + open_fret_width,
            self.config.general.y_start,
        )
        end = (
            self.config.general.x_start + open_fret_width,
            self.config.general.y_start
            + self.config.general.fret_height * (len(self.tuning) - 1),
        )
        nut = Nut(start, end, config=self.config.nut)
        self.elements.nut = nut

    def add_fret_numbers(self) -> None:
        """Build and add fret number elements."""
        if not self.config.general.show_frets:
            return None

        dots = get_valid_dots(
            self.config.general.first_fret, self.config.general.last_fret
        )
        for dot in dots:
            x = self.config.general.x_start + self.config.general.fret_width * (
                1 / 2 + dot - self.config.general.first_fret
            )
            y = self.config.general.y_start + self.config.general.fret_height * (
                len(self.tuning)
            )
            fret_number = FretNumber(str(dot), (x, y), config=self.config.fret_numbers)
            self.elements.fret_numbers.append(fret_number)

    def add_tuning(self) -> None:
        """Build and add tuning element."""
        if not self.config.general.show_tuning:
            return None
        reversed_tuning = list(reversed(self.tuning))
        for i, note in enumerate(reversed_tuning):
            x = self.config.general.x_start + (
                self.config.general.fret_width
                * (
                    self.config.general.last_fret
                    - self.config.general.first_fret
                    + 3 / 2
                )
            )
            y = self.config.general.y_start + self.config.general.fret_height * (i)
            tuning_note = Tuning(note, (x, y), config=self.config.tuning)
            self.elements.tuning.append(tuning_note)

    def _get_open_note(
        self, position: Tuple[float, float], note: str, root: Optional[str] = None
    ) -> OpenNote:
        config = copy.copy(self.config.open_notes)
        if root and self.config.general.open_color_scale:
            idx = chromatic_position_from_root(note, root)
            color = self.config.general.open_colors.from_interval(idx)
            config.color = color

        if root and self.config.general.show_degree_name:
            note = note_to_interval_name(note, root)

        if not (
            self.config.general.show_degree_name or self.config.general.show_note_name
        ):
            note = ""

        _note = OpenNote(note, position, config=config)
        return _note

    def _get_fretted_note(
        self, position: Tuple[float, float], note: str, root: Optional[str] = None
    ) -> FrettedNote:
        config = copy.copy(self.config.fretted_notes)
        if root and self.config.general.fretted_color_scale:
            idx = chromatic_position_from_root(note, root)
            color = self.config.general.fretted_colors.from_interval(idx)
            config.color = color

        if root and self.config.general.show_degree_name:
            note = note_to_interval_name(note, root)

        if not (
            self.config.general.show_degree_name or self.config.general.show_note_name
        ):
            note = ""

        _note = FrettedNote(note, position, config=config)
        return _note

    def add_note(self, string_no: int, note: str, root: Optional[str] = None) -> None:
        """Build and add background element."""
        if string_no < 0 or string_no > len(self.tuning):
            raise ValueError(f"String number is invalid. Tuning is {self.tuning}")

        string_root = self.tuning[len(self.tuning) - 1 - string_no]
        _idx = chromatic_position_from_root(note, string_root)

        indices = []
        while _idx <= self.config.general.last_fret:
            indices.append(_idx)
            _idx += 12

        for idx in indices:
            x = self.config.general.x_start + (self.config.general.fret_width) * (
                idx + (1 / 2)
            )
            y = self.config.general.y_start + self.config.general.fret_height * (
                string_no
            )
            position = (x, y)
            _note: Union[OpenNote, FrettedNote]
            if idx == 0:
                _note = self._get_open_note(position, note, root)
            else:
                _note = self._get_fretted_note(position, note, root)
            self.elements.notes.append(_note)

    def add_notes(self, scale: NotesContainer) -> None:
        """Add an entire scale (from NoteContainer object) to the fretboard.

        Parameters
        ----------
        scale : NotesContainer
            Object representing the root and the associated scale
        """
        for string_no, _ in enumerate(self.tuning):
            for note in scale.notes:
                self.add_note(string_no, note, scale.root)

    def add_fingering(
        self, fingering: List[Optional[int]], root: Optional[str] = None
    ) -> None:
        """Add fingering starting with upper string to lower string.

        This function automatically calculate the note names or
        intervals if the root is given Display them in the way described
        in configuration
        """
        if len(fingering) != len(self.tuning):
            raise ValueError(
                f"Fingering has not the same size as tuning."
                f" Got {len(fingering)} expected {len(self.tuning)}"
            )

        for string_no, finger_position in enumerate(reversed(fingering)):
            string_note = self.tuning[len(self.tuning) - 1 - string_no]
            if finger_position is None:
                x = self.config.general.x_start + self.config.general.fret_width * (
                    1 / 2
                )
                y = self.config.general.y_start + (self.config.general.fret_height) * (
                    string_no
                )
                position = (x, y)
                cross = Cross(position, config=self.config.cross)
                self.elements.crosses.append(cross)
            else:
                x = self.config.general.x_start + self.config.general.fret_width * (
                    1 / 2 + finger_position
                )
                y = self.config.general.y_start + (self.config.general.fret_height) * (
                    string_no
                )

                position = (x, y)
                note_name = get_note_from_index(finger_position, string_note)

                _note: Union[OpenNote, FrettedNote]
                if finger_position == 0:
                    _note = self._get_open_note(position, note_name, root)
                else:
                    _note = self._get_fretted_note(position, note_name, root)

                self.elements.notes.append(_note)

    def get_size(self) -> Tuple[float, float]:
        """Get total size of the drawing.

        Returns
        -------
        Tuple[float, float]
            Width and heigth
        """
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        width = (
            self.config.general.x_start
            + self.config.general.fret_width * (number_of_frets + 2)
            + self.config.general.x_end_offset
        )
        height = (
            self.config.general.y_start
            + self.config.general.fret_height * (len(self.tuning) + 1)
            + self.config.general.y_end_offset
        )
        return (width, height)

    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Get the size of the inner drawing.

        This function could be use to add custom elements

        Returns
        -------
        Tuple[Tuple[float, float], Tuple[float, float]]
            Upper left corner x coordinate, upper left corner y coordinate
            Lower right corner x coordinate, lower right corner y coordinate
        """
        open_fret_width = self.config.general.fret_width
        number_of_frets = self.config.general.last_fret - self.config.general.first_fret
        upper_left_x = self.config.general.x_start
        upper_left_y = self.config.general.y_start
        lower_right_x = (
            self.config.general.x_start
            + open_fret_width
            + (number_of_frets) * (self.config.general.fret_width)
        )
        lower_right_y = (
            self.config.general.y_start
            + (len(self.tuning) - 1) * self.config.general.fret_height
        )
        return ((upper_left_x, upper_left_y), (lower_right_x, lower_right_y))

    def add_note_element(self, note: Union[OpenNote, FrettedNote]) -> None:
        """Add a note element to the fretboard.

        Need to be either a OpenNote or a FrettedNote

        Parameters
        ----------
        note : Union[OpenNote, FrettedNote]
            Note element

        Raises
        ------
        ValueError
            If the note is not a Union[OpenNote, FrettedNote]
        """
        self.elements.notes.append(note)

    def add_element(self, element: FretBoardElement) -> None:
        """Add an element to the fretboard.

        Need to be either a FretBoardElement

        Parameters
        ----------
        element : FretBoardElement
            Fretboard element

        Raises
        ------
        ValueError
            If the element is not a FretboardElement
        """
        if not issubclass(type(element), FretBoardElement):
            raise ValueError("Element should be a 'FretBoardElement'")
        self.elements.customs.append(element)

    def get_elements(self) -> FretBoardElements:
        return self.elements



from __future__ import absolute_import

from enum import Enum


class Mode(str, Enum):
    """Makes it easier to list and select a mode.

    One can use auto-completion Mode.MIXOLYDIAN can be used instead of
    the 'Mixolydian' literal string.
    """

    AEOLIAN = "Aeolian"
    ALTERED = "Altered"
    AUGMENTED_LYDIAN = "Augmentedlydian"
    DOMINANT_BEBOP = "Dominantbebop"
    DORIAN = "Dorian"
    DORIAN_B9 = "Dorianb9"
    DORIAN_SHARP11 = "Doriansharp11"
    HALF_TONE_WHOLE_TONE = "Halftonewholetone"
    HARMONIC_MINOR = "Harmonicminor"
    IONIAN = "Ionian"
    IONIAN_SHARP5 = "Ioniansharp5"
    LOCRIAN = "Locrian"
    LOCRIAN_BEC13 = "Locrianbec13"
    LOCRIAN_BEC9 = "Locrianbec9"
    LYDIAN = "Lydian"
    LYDIAN_B7 = "Lydianb7"
    LYDIAN_BEC9 = "Lydianbec9"
    MAJOR = "Ionian"  # For convenience
    MAJOR_BEBOP = "Majorbebop"
    MAJOR_BLUE = "Majorblue"
    MAJOR_PENTATONIC = "Majorpentatonic"
    MELOD_ICMINOR = "Melodicminor"
    MINOR_BLUES = "Minorblues"
    MINOR_PENTATONIC = "Minorpentatonic"
    MIXOLYDIAN = "Mixolydian"
    MIXOLYDIAN_B13 = "Mixolydianb13"
    MIXOLYDIAN_B9_B13 = "Mixolydianb9b13"
    NATURAL_MINOR = "Aeolian"
    PHRYGIAN = "Phrygian"
    SUPER_LOCRIAN_BB7 = "Superlocrianbb7"
    WHOLE_TONE = "Wholetone"
    WHOLE_TONE_HALF_TONE = "Wholetonehalftone"


SCALES_DICT = {
    "Ionian": [0, 2, 4, 5, 7, 9, 11],
    "Dorian": [0, 2, 3, 5, 7, 9, 10],
    "Phrygian": [0, 1, 3, 5, 7, 8, 10],
    "Lydian": [0, 2, 4, 6, 7, 9, 11],
    "Mixolydian": [0, 2, 4, 5, 7, 9, 10],
    "Aeolian": [0, 2, 3, 5, 7, 8, 10],
    "Locrian": [0, 1, 3, 5, 6, 8, 10],
    "Melodicminor": [0, 2, 3, 5, 7, 9, 11],
    "Dorianb9": [0, 1, 3, 5, 7, 9, 10],
    "Augmentedlydian": [0, 2, 4, 6, 8, 9, 11],
    "Lydianb7": [0, 2, 4, 6, 7, 9, 10],
    "Mixolydianb13": [0, 2, 4, 5, 7, 8, 10],
    "Locrianbec9": [0, 2, 3, 5, 6, 8, 10],
    "Altered": [0, 1, 3, 4, 6, 8, 10],
    "Harmonicminor": [0, 2, 3, 5, 7, 8, 11],
    "Locrianbec13": [0, 1, 3, 5, 6, 9, 10],
    "Ioniansharp5": [0, 2, 4, 5, 8, 9, 11],
    "Doriansharp11": [0, 2, 3, 6, 7, 9, 10],
    "Mixolydianb9b13": [0, 1, 4, 5, 7, 8, 10],
    "Lydianbec9": [0, 3, 4, 6, 7, 9, 11],
    "Superlocrianbb7": [0, 1, 3, 4, 6, 8, 9],
    "Wholetone": [0, 2, 4, 6, 8, 10],
    "Majorpentatonic": [0, 2, 4, 7, 9],
    "Minorpentatonic": [0, 3, 5, 7, 10],
    "Majorblue": [0, 2, 3, 4, 7, 9],
    "Minorblues": [0, 3, 5, 6, 7, 10],
    "Halftonewholetone": [0, 1, 3, 4, 6, 7, 9, 10],
    "Wholetonehalftone": [0, 2, 3, 5, 6, 8, 9, 11],
    "Dominantbebop": [0, 2, 4, 5, 7, 9, 10, 11],
    "Majorbebop": [0, 2, 3, 5, 7, 8, 9, 10],
}


class Chord(str, Enum):
    """Makes it easier to list and select a chord.

    One can use auto-completion Chord.MAJOR can be used instead of the 'M'
    literal string.

    Not every defined Chord from CHORDS_DICT_ESSENTIAL is defined as an
    enum here, only the most common ones.
    """

    MAJOR = "M"
    ELEVENTH = "11"
    THIRTEENTH = "13"
    FIFTH = "5"
    POWER = "5"  # For convenience
    SIXTH = "6"
    SEVENTH = "7"
    DOMINANT_SEVENTH = "7"  # For convenience
    NINTH = "9"
    AUGMENTED = "aug"
    DIMINISHED = "dim"
    DIMINISHED_SEVENTH = "dim7"
    MINOR = "m"
    MINOR_ELEVENTH = "m11"
    MINOR_THIRTEENTH = "m13"
    MINOR_SIXTH = "m6"
    MINOR_SEVENTH = "m7"
    MINOR_NINTH = "m9"
    MAJOR_SEVENTH = "maj7"
    SUSPENDED_FOURTH = "sus4"
    SUSPENDED_SECOND = "sus2"


CHORDS_DICT_ESSENTIAL = {
    "M": [0, 4, 7],
    "(b5)": [0, 4, 6],
    "11": [0, 4, 5, 7, 10],
    "11#5": [0, 4, 5, 8, 10],
    "11b5": [0, 4, 5, 6, 10],
    "13": [0, 4, 7, 10, 9],
    "13#5": [0, 4, 8, 9, 10],
    "13b5": [0, 4, 6, 9, 10],
    "5": [0, 7],
    "6": [0, 4, 7, 9],
    "6b5": [0, 4, 6, 9],
    "7": [0, 4, 7, 10],
    "7#5": [0, 4, 8, 10],
    "7b5": [0, 4, 6, 10],
    "7sus2": [0, 2, 7, 10],
    "7sus4": [0, 5, 7, 10],
    "9": [0, 4, 7, 10, 2],
    "9#5": [0, 4, 8, 10, 2],
    "9b5": [0, 4, 6, 10, 2],
    "9sus2": [0, 2, 7, 10, 2],
    "9sus4": [0, 5, 7, 10, 2],
    "aug": [0, 4, 8],
    "aug6": [0, 4, 8, 9],
    "dim": [0, 3, 6],
    "dim(maj11)": [0, 3, 5, 6, 11],
    "dim(maj13)": [0, 3, 6, 9, 11],
    "dim(maj7)": [0, 3, 6, 11],
    "dim(maj9)": [0, 2, 3, 6, 11],
    "dim11": [0, 3, 5, 6, 10],
    "dim13": [0, 3, 6, 9, 10],
    "dim6": [0, 3, 6, 9],
    "dim7": [0, 3, 6, 9],
    "dim9": [0, 2, 3, 6, 10],
    "m": [0, 3, 7],
    "m#5": [0, 3, 8],
    "m(maj11)": [0, 3, 5, 7, 11],
    "m(maj11)#5": [0, 3, 5, 8, 11],
    "m(maj13)": [0, 3, 7, 9, 11],
    "m(maj13)#5": [0, 3, 8, 9, 11],
    "m(maj7)": [0, 3, 7, 11],
    "m(maj7)#5": [0, 3, 8, 11],
    "m(maj9)": [0, 2, 3, 7, 11],
    "m(maj9)#5": [0, 2, 3, 8, 11],
    "m11": [0, 3, 7, 10, 5],
    "m11#5": [0, 3, 5, 8, 10],
    "m13": [0, 3, 7, 10, 9],
    "m13#5": [0, 3, 8, 9, 10],
    "m6": [0, 3, 7, 9],
    "m6#5": [0, 3, 8, 9],
    "m7": [0, 3, 7, 10],
    "m7#5": [0, 3, 8, 10],
    "m7b5": [0, 3, 6, 10],
    "m9": [0, 3, 7, 10, 2],
    "m9#5": [0, 2, 3, 8, 10],
    "maj11": [0, 4, 5, 7, 11],
    "maj11#5": [0, 4, 5, 8, 11],
    "maj11b5": [0, 4, 5, 6, 11],
    "maj13": [0, 4, 7, 11, 9],
    "maj13#5": [0, 4, 8, 9, 11],
    "maj13b5": [0, 4, 6, 9, 11],
    "maj7": [0, 4, 7, 11],
    "maj7#5": [0, 4, 8, 11],
    "maj7b5": [0, 4, 6, 11],
    "maj9": [0, 4, 7, 11, 2],
    "maj9#5": [0, 2, 4, 8, 11],
    "maj9b5": [0, 2, 4, 6, 11],
    "sus2": [0, 2, 7],
    "sus2(#5)": [0, 2, 8],
    "sus2(b5)": [0, 2, 6],
    "sus4": [0, 5, 7],
    "sus4(#5)": [0, 5, 8],
    "sus4(b5)": [0, 5, 6],
}


class Note(str, Enum):
    """Makes it easier to list and select a note.

    One can use auto-completion Note.A_SHARP instead of the 'A#' literal
    string.
    """

    A = "A"
    B = "B"
    C = "C"
    D = "D"
    E = "E"
    F = "F"
    G = "G"
    A_SHARP = "A#"
    A_FLAT = "Ab"
    B_FLAT = "Bb"
    C_FLAT = "Cb"
    C_SHARP = "C#"
    D_FLAT = "Db"
    D_SHARP = "D#"
    E_FLAT = "Eb"
    E_SHARP = "E#"
    F_SHARP = "F#"
    G_FLAT = "Gb"
    G_SHARP = "G#"


class Interval(str, Enum):
    """Makes it easier to list and select an interval.

    One can use auto-completion Interval.DIMINISHED_FIFTH instead of the
    'b5' literal string.
    """

    ROOT = "1"
    MINOR_SECOND = "b2"
    MAJOR_SECOND = "2"
    MINOR_THIRD = "b3"
    MAJOR_THIRD = "3"
    PERFECT_FOURTH = "4"
    DIMINISHED_FIFTH = "b5"
    PERFECT_FIFTH = "5"
    MINOR_SIXTH = "b6"
    MAJORS_IXTH = "6"
    MINOR_SEVENTH = "b7"
    MAJOR_SEVENTH = "7"


class LongInterval(str, Enum):
    """Makes it easier to list and select a long interval.

    One can use LongInterval.MINOR_SECOND instead of the 'minor_second'
    literal string.
    """

    ROOT = "root"
    MINOR_SECOND = "minor_second"
    MAJOR_SECOND = "major_second"
    MINOR_THIRD = "minor_third"
    MAJOR_THIRD = "major_third"
    PERFECT_FOURTH = "perfect_fourth"
    DIMINISHED_FIFTH = "diminished_fifth"
    PERFECT_FIFTH = "perfect_fifth"
    MINOR_SIXTH = "minor_sixth"
    MAJORS_IXTH = "major_sixth"
    MINOR_SEVENTH = "minor_seventh"
    MAJOR_SEVENTH = "major_seventh"


CHROMATICS_NOTES = ["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"]
CHROMATICS_INTERVALS = ["1", "b2", "2", "b3", "3", "4", "b5", "5", "b6", "6", "b7", "7"]

ALTERATIONS = {
    "Bb": "A#",
    "Db": "C#",
    "Eb": "D#",
    "Gb": "F#",
    "Ab": "G#",
}

SHARPY_ALTERATIONS = {value: key for (key, value) in ALTERATIONS.items()}
ENHARMONICS = {
    "A#": "Bb",
    "B": "Cb",
    "F": "E#",
    "C": "B#",
    "B#": "C",
    "C#": "Db",
    "D#": "Eb",
    "E": "Fb",
    "F#": "Gb",
    "G#": "Ab",
}

INTERVAL_MAPPING = {
    "root": "1",
    "minorsecond": "b2",
    "majorsecond": "2",
    "minorthird": "b3",
    "majorthird": "3",
    "perfectfourth": "4",
    "diminishedfifth": "b5",
    "perfectfifth": "5",
    "minorsixth": "b6",
    "majorsixth": "6",
    "minorseventh": "b7",
    "majorseventh": "7",
}

DOTS_FRETBOARD_POSITIONS = [0, 3, 5, 7, 9, 12]
NUMBER_OF_DOTS = [2, 1, 1, 1, 1, 2]

STANDARD_TUNING = ["E", "A", "D", "G", "B", "E"]

WHITE = "rgb(255,255,255)"
BLACK = "rgb(0,0,0)"
GRAY = "rgb(150,150,150)"
DARK_GRAY = "rgb(200,200,200)"
NO_COLOR = "none"

DOTS_POSITIONS = [3, 5, 7, 9, 12, 15, 17, 19, 21, 24]



from typing import Collection, List

from fretboardgtr.constants import (
    ALTERATIONS,
    CHROMATICS_INTERVALS,
    CHROMATICS_NOTES,
    DOTS_POSITIONS,
    ENHARMONICS,
    SHARPY_ALTERATIONS,
)


def get_valid_dots(first_fret: int, last_fret: int) -> List[int]:
    """Get the valid fretboard dots between frets.

    Parameters
    ----------
    first_fret : int
        First fret
    last_fret : int
        Last fret

    Returns
    -------
    List[int]
        Valid dots between frets
    """
    dots = []
    for dot in DOTS_POSITIONS:
        if dot >= first_fret and dot <= last_fret:
            dots.append(dot)
    return dots


def _contains_duplicates(iterable: Collection) -> bool:
    if len(iterable) != len(set(iterable)):
        return True
    return False


def chromatics_from_root(root: str) -> List[str]:
    """Create list of notes chromatically starting with root."""
    notes = []
    root_idx = CHROMATICS_NOTES.index(root)
    for i, _ in enumerate(CHROMATICS_NOTES):
        notes.append(CHROMATICS_NOTES[(root_idx + i) % 12])
    return notes


def get_note_from_index(index: int, root: str) -> str:
    """Get note from chromatic scale from index and root."""
    chroma_from_root = chromatics_from_root(root)
    return chroma_from_root[index % 12]


def chromatic_position_from_root(note: str, root: str) -> int:
    """Get the index of the note from the root on chromatic scale."""
    idx = 0
    chroma_from_root = chromatics_from_root(root)
    for _idx, chromatic_note in enumerate(chroma_from_root):
        if chromatic_note == note:
            idx = _idx
    return idx


def to_sharp_note(note: str) -> str:
    """Get note by replacing it by its corresponding sharp note."""
    if note in ALTERATIONS:
        note = ALTERATIONS[note]
    return note


def to_flat_note(note: str) -> str:
    """Get note by replacing it by its corresponding flat note."""
    if note in SHARPY_ALTERATIONS:
        note = SHARPY_ALTERATIONS[note]
    return note


def scale_to_sharp(scale: List[str]) -> List[str]:
    """Get scale replacing each note by its sharp correspondant note."""
    flat_scale = list(scale)
    for i, note in enumerate(scale):
        sharp_note = to_sharp_note(note)
        flat_scale[i] = sharp_note
    return flat_scale


def scale_to_flat(scale: List[str]) -> List[str]:
    """Get scale replacing each note by its flat correspondant note."""
    flat_scale = list(scale)
    for i, note in enumerate(scale):
        sharp_note = to_flat_note(note)
        flat_scale[i] = sharp_note
    return flat_scale


def note_to_interval(note: str, root: str) -> int:
    """Get note from interval (int)."""
    sharp_note = to_sharp_note(note)
    idx = chromatic_position_from_root(sharp_note, root)
    return idx


def note_to_interval_name(note: str, root: str) -> str:
    """Get note from interval name."""
    idx = note_to_interval(note, root)
    return CHROMATICS_INTERVALS[idx]


def scale_to_intervals(scale: List[str], root: str) -> List[int]:
    """Get intervals from root.

    >>> scale_to_intervals(scale=['C','E','G'],root='C')
    [1,3,5]
    """
    intervals = []
    for note in scale:
        intervals.append(note_to_interval(note, root))
    return intervals


def scale_to_enharmonic(scale: List[str]) -> List[str]:
    """Modify the scale in order to not repeat note.

    Turns into enharmonic way if possible.

    >>> setenharmonic(['A#', 'C', 'D', 'D#', 'F', 'G', 'A'])
        ['Bb', 'C', 'D', 'Eb', 'F', 'G', 'A']
    >>> setenharmonic(["A","A#","B","C","C#","D","D#","E","F","F#","G","G#"])
        ['A', 'A#', 'B', 'C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#']
    >>> setenharmonic(['F#', 'G#', 'A#', 'B', 'C#', 'D#', 'F'])
        ['Gb', 'Ab', 'Bb', 'Cb', 'Db', 'Eb', 'F']
    """
    sharp_scale = scale_to_sharp(scale)
    enharmonic_scale = list(sharp_scale)

    if _contains_duplicates(sharp_scale):
        for i, note in enumerate(sharp_scale):
            # len == 2 means there is an alteration
            if len(note) == 2:
                enharmonic_scale[i] = ENHARMONICS[note]

    # If there is still duplicate then remove transform B and F
    if _contains_duplicates(enharmonic_scale):
        if "B" in enharmonic_scale:
            enharmonic_scale[enharmonic_scale.index("B")] = ENHARMONICS["B"]
        elif "F" in enharmonic_scale:
            enharmonic_scale[enharmonic_scale.index("F")] = ENHARMONICS["F"]

    return enharmonic_scale



from dataclasses import fields

import svgwrite

from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.fretboards.base import FretBoardLike


class FretBoardToSVGConverter:
    """Convert a FretboardLike object to a svgwrite object.

    Convert it in order to export it to a specific format later.
    """

    def __init__(self, fretboard: FretBoardLike):
        self._fretboard = fretboard
        self.drawing = self.get_empty()

    def get_empty(self) -> svgwrite.Drawing:
        """Create empty box and the object self.drawing."""
        width, height = self._fretboard.get_size()
        return svgwrite.Drawing(
            size=(width, height),
            profile="full",
        )

    def add_to_drawing(
        self, drawing: svgwrite.Drawing, element: FretBoardElement
    ) -> svgwrite.Drawing:
        if not issubclass(type(element), FretBoardElement):
            raise ValueError(f"Element {element} does not subclass FretBoardElement")
        drawing.add(element.get_svg())
        return drawing

    def convert(self) -> svgwrite.Drawing:
        drawing = self.get_empty()
        elements = self._fretboard.get_elements()
        for key in fields(elements):
            element = getattr(elements, key.name, None)
            if isinstance(element, list):
                for sub in element:
                    drawing = self.add_to_drawing(drawing, sub)
            else:
                drawing = self.add_to_drawing(drawing, element)  # type: ignore
        return drawing



from dataclasses import dataclass
from typing import List

from fretboardgtr.constants import CHORDS_DICT_ESSENTIAL, CHROMATICS_NOTES, SCALES_DICT


@dataclass
class NotesContainer:
    root: str
    notes: List[str]


class ScaleFromName:
    """Object that generating NotesContainer object from root and mode.

    Given a root name and a mode name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.Mode enum as well as string
    Note name can be given thanks to the constants.Note enum as well as string

    Example
    -------
    >>> ScaleFromName(root='C',mode='Dorian').get()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    >>> ScaleFromName(root=Note.C,mode=Mode.DORIAN).get()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    """

    def __init__(self, root: str = "C", mode: str = "Ionian"):
        self.root = root
        self.mode = mode

    def get(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)
        mode_idx = SCALES_DICT[self.mode]
        scale = []
        for note_id in mode_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)


class ChordFromName:
    """Object generating NotesContainer object from root and chord quality.

    Given a root name and a quality name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.Chord enum as well as string
    Note name can be given thanks to the constants.Note enum as well as string

    Example
    -------
    >>> ScaleFromName(root='C',quality='M).get()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    >>> ScaleFromName(root=Note.C,quality=Chord.MAJOR).resultget()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    """

    def __init__(self, root: str = "C", quality: str = "M"):
        self.root = root
        self.quality = quality

    def get(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)

        quality_idx = CHORDS_DICT_ESSENTIAL[self.quality]
        scale = []
        for note_id in quality_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)



from fretboardgtr.notes_creators import ChordFromName, ScaleFromName


def test_scale_creator():
    scale = ScaleFromName(root="C", mode="Ionian").get()
    print(scale.notes)
    assert scale.notes == ["C", "D", "E", "F", "G", "A", "B"]


def test_chord_creator():
    chord = ChordFromName(root="C", quality="M").get()
    assert chord.notes == ["C", "E", "G"]



"""The setup script."""
from typing import Dict

from setuptools import setup

version: Dict[str, str] = {}
with open("fretboardgtr/_version.py") as fp:
    exec(fp.read(), version)

setup(version=version["version_str"])



from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

from fretboardgtr.elements.base import FretBoardElement
from fretboardgtr.elements.notes import FrettedNote, OpenNote
from fretboardgtr.exporters import EXPORTERS
from fretboardgtr.fretboards.base import FretBoardLike
from fretboardgtr.fretboards.converters import FretBoardToSVGConverter
from fretboardgtr.fretboards.fretboard import FretBoardConfig, FretBoardContainer
from fretboardgtr.fretboards.vertical import VerticalFretBoardContainer
from fretboardgtr.notes_creators import NotesContainer


class FretBoard:
    """High level API class for fretboard generation."""

    def __init__(
        self,
        tuning: Optional[List[str]] = None,
        config: Optional[FretBoardConfig] = None,
    ):
        self._fretboard: FretBoardLike = FretBoardContainer(
            config=config, tuning=tuning
        )

    def set_config(self, config: FretBoardConfig) -> None:
        """Set a new configuration for the fretboard.

        Changing the configuration does not affect elements already
        defined. Only new ones will be changed. If you want to change
        completely the fretboard from configuration, a new object should
        be created
        """
        if not issubclass(type(config), FretBoardConfig):
            raise ValueError(f"scale should be a FretBoardConfig nor {type(config)}")
        self._fretboard.set_config(config)

    def add_notes(self, scale: NotesContainer) -> None:
        """Add an entire scale (from NoteContainer object) to the fretboard.

        See :func:`~fretboardgtr.fretboards.fretboard.add_notes`
        """
        if not issubclass(type(scale), NotesContainer):
            raise ValueError(f"scale should be a NotesContainer nor {type(scale)}")
        self._fretboard.add_notes(scale=scale)

    def add_fingering(self, fingering: List[Optional[int]]) -> None:
        """Add fingering starting with upper string to lower string.

        See :func:`~fretboardgtr.fretboards.fretboard.add_fingering`
        """
        self._fretboard.add_fingering(fingering=fingering)

    def add_note_element(self, note: Union[OpenNote, FrettedNote]) -> None:
        """Add a note element to the fretboard.

        See :func:`~fretboardgtr.fretboards.fretboard.add_note_element`
        """
        self._fretboard.add_note_element(note)

    def add_element(self, element: FretBoardElement) -> None:
        """Add an element to the fretboard.

        See :func:`~fretboardgtr.fretboards.fretboard.add_element`
        """
        self._fretboard.add_element(element)

    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Get inside bounds of the fretboard.

        See :func:`~fretboardgtr.fretboards.fretboard.get_inside_bounds`
        """
        return self._fretboard.get_inside_bounds()

    def export(
        self, to: Union[Path, str], format: Literal["svg", "png", "pdf"] = "svg"
    ) -> None:
        """Export the fretboard to the desired path.

        Export availables are :
            svg
            png
            pdf

        Custom exporters can be created by :
        Subclassing : func:`~fretboardgtr.exporters.Exporter`
        Regitering it with func:`~fretboardgtr.exporters.register_exporter`

        Parameters
        ----------
        to : Union[Path, str]
            Path to export to
        format : Literal[&quot;svg&quot;, &quot;png&quot;, &quot;pdf&quot;], optional
            Format to export in, by default "svg"

        Raises
        ------
        NotImplementedError
            If the desired format is not implemented
        """
        if format.upper() not in EXPORTERS:
            availables = ", ".join(list(EXPORTERS.keys()))
            raise NotImplementedError(
                f"Save to the {format} format is unsupported."
                f" Available formats are {availables}"
            )
        drawing = FretBoardToSVGConverter(self._fretboard).convert()
        EXPORTERS[format.upper()](drawing).export(to=to)


class VerticalFretBoard(FretBoard):
    def __init__(
        self,
        tuning: Optional[List[str]] = None,
        config: Optional[FretBoardConfig] = None,
    ):
        self._fretboard: FretBoardLike = VerticalFretBoardContainer(
            config=config, tuning=tuning
        )

    def add_notes(self, scale: NotesContainer) -> None:
        """Add an entire scale (from NoteContainer object) to the fretboard.

        See :func:`~fretboardgtr.fretboards.vertical.add_notes`
        """
        super().add_notes(scale)

    def add_fingering(self, fingering: List[Optional[int]]) -> None:
        """Add fingering starting with upper string to lower string.

        See :func:`~fretboardgtr.fretboards.vertical.add_fingering`
        """
        super().add_fingering(fingering=fingering)

    def add_note_element(self, note: Union[OpenNote, FrettedNote]) -> None:
        """Add a note element to the fretboard.

        See :func:`~fretboardgtr.fretboards.vertical.add_note_element`
        """
        super().add_note_element(note)

    def add_element(self, element: FretBoardElement) -> None:
        """Add an element to the fretboard.

        See :func:`~fretboardgtr.fretboards.vertical.add_element`
        """
        super().add_element(element)

    def get_inside_bounds(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Get inside bounds of the fretboard.

        See :func:`~fretboardgtr.fretboards.vertical.get_inside_bounds`
        """
        return super().get_inside_bounds()
